#!/usr/bin/python
# Copyright (C) 2008 One Laptop Per Child Association, Inc.
# Licensed under the terms of the GNU GPL v2 or later; see COPYING for details.
# Written by C. Scott Ananian <cscott@laptop.org>
"""Extensions to the sugar.bundle.activitybundle package, and resources for
discovering update sources."""
from __future__ import with_statement
import os, os.path, re, traceback, urllib, urlparse
import bitfrost.util.urlrange as urlrange

import sugar.bundle.activitybundle as actbun
import sugar.bundle.contentbundle as libbun
import sugar.env as env

USER_ACTIVITY_DIR=env.get_user_activities_path()
USER_LIBRARY_DIR=env.get_user_library_path()
USER_GROUPS_FILE=os.path.join(USER_ACTIVITY_DIR, '.groups')
SYSTEM_ACTIVITY_DIR='/usr/share/sugar/activities'
SYSTEM_LIBRARY_DIR='/usr/share/sugar/activities'
SYSTEM_GROUPS_FILE='/etc/olpc-update/activity-groups'

DEFAULT_UPDATE_URL = "http://wiki.laptop.org/go/Activities"
"""For backwards-compatibility, hard-code a default update URLs for
already-distributed bundles that don't (yet) have them."""
DEFAULT_G1G1_GROUP = '%s/G1G1Lite' % DEFAULT_UPDATE_URL
"""For backwards-compatibility with pre-703 builds, add a G1G1 group
by default."""

_DEBUG_VERSIONS=False
"""Hard-code sample version values for debugging."""

DEFAULT_LIBRARY_ICON=\
"""<?xml version="1.0" ?><!DOCTYPE svg  PUBLIC '-//W3C//DTD SVG 1.1//EN'  'http://www.w3.org/Graphics/SVG/1.1/DTD/svg11.dtd' [
	<!ENTITY stroke_color "#000">
	<!ENTITY fill_color "#eee">
]><svg height="55px" viewBox="0 0 55 55" width="55px" xmlns="http://www.w3.org/2000/svg" xmlns:xlink="http://www.w3.org/1999/xlink" stroke-linecap="round" stroke-linejoin="round" stroke-width="3.5" stroke="&stroke_color;" fill="&fill_color;">
 <g>
  <path d="M 27.904 11.023 L 27.902 11.023 C 27.902 11.021 26.192 8.97 18.526 8.519 C 10.86 8.07 6.843 10.121 6.84 10.122 C 4.942 10.741 3.345 11.857 3.345 13.616 L 3.345 41.318 C 3.345 43.343 4.58 44.812 6.84 44.812 C 6.843 44.812 11.25 43.462 16.844 43.462 C 22.433 43.461 27.905 45.715 27.905 45.715 " />
  <path d="M 27.898 11.023 L 27.898 11.023 C 27.898 11.021 29.613 8.97 37.275 8.519 C 44.943 8.07 48.961 10.121 48.963 10.122 C 50.86 10.741 52.457 11.857 52.457 13.616 L 52.457 41.318 C 52.457 43.343 51.224 44.812 48.963 44.812 C 48.96 44.812 44.554 43.462 38.959 43.462 C 33.37 43.461 27.897 45.715 27.897 45.715 " />
  <path d="M 27.898 11.023 L 27.9 45.717 " fill="none" />
  <path d="M 32.566 44.275 C 32.566 44.275 32.535 47.181 27.9 47.181 C 23.268 47.181 23.237 44.275 23.237 44.275 " fill="none" />
 </g><!--"-->
</svg>"""
"""Hack in an icon for library content bundles."""

def _lookup_update_url(ab, is_activity=True):
    # reparse configuration file.
    from ConfigParser import ConfigParser
    cp = ConfigParser()
    if is_activity:
        cp.readfp(ab.get_file('activity/activity.info'))
        section = 'Activity'
    else:
        cp.readfp(ab.get_file('library/library.info'))
        section = 'Library'
    # we've got a default update_url if none is specified
    ab._update_url = DEFAULT_UPDATE_URL
    # if an explicit update_url is listed, use that instead.
    if cp.has_option(section, 'update_url'):
        ab._update_url = cp.get(section, 'update_url')

def _lookup_global_name(lb):
    # reparse library.info file.
    from ConfigParser import ConfigParser
    cp = ConfigParser()
    cp.readfp(lb.get_file('library/library.info'))
    section = 'Library'
    # this is required.
    lb._global_name = cp.get(section, 'global_name')

def _get_icon_data(ab):
    fname = ab.get_icon()
    try:
        with open(fname) as f:
            return f.read()
    finally:
        if ab._zip_file is not None: # dirty dirty hack
            os.unlink(fname)

def get_major_version():
    """Return the first two components of a dotted OLPC release number."""
    relnum = get_release_version()
    if relnum is None: return None # bail!
    major, nsubs = re.subn(r'^(\d+[.]\d+)[.]\d+$', r'\1', relnum, 1)
    return major if nsubs == 1 else None

def get_release_version():
    """Return the OLPC release number, which is a two- or three-part
    dotted version number like, "8.1"."""
    if _DEBUG_VERSIONS: return "8.1.1"
    try:
        return open('/etc/olpc-release').read().strip()
    except:
        return None
def get_os_version():
    """Return the "core OS" version, which is a integer optionally
    prefixed with a stream name, like "656" or "joyride 2100"."""
    if _DEBUG_VERSIONS: return "joyride 2100"
    try:
        return open('/boot/olpc_build').read().strip()
    except:
        return None

def url_variants(url):
    """Return a list of variants for a given URL, representing release
    version dependencies."""
    suffixes = [get_os_version(), get_release_version(), get_major_version()]
    return [urlparse.urljoin(url+'/', urllib.quote(s))
            for s in suffixes if s is not None] + [ url ]

def retrieve_first_variant(url, parse_func=(lambda f, u: f.read()),
                           **urlopen_args):
    for u in url_variants(url):
        try:
            uo = urlrange.urlopen(u, **urlopen_args)
            result = parse_func(uo, uo.url)
            if result is not None: return result
        except:
            continue # some failure, go on to next possible url
    return None # apparently, none of these URLs worked today.

# activity groups are represented by ~olpc/Activities/.groups
# which consists of a simple list of urls, one url per line.
def get_activity_group_urls():
    """Return a list of "group urls", which are the locations at which we
    can find group definitions.  Try a file in the home dir, then fall
    back to a build default, if present."""
    for filename in [ USER_GROUPS_FILE, SYSTEM_GROUPS_FILE ]:
        try:
            return [line.strip() for line in open(filename)
                    if not (line.strip()=='' or line.startswith('#'))]
        except IOError:
            continue # try next file.
    # hmm, no groups file.  let's make a special test to support g1g1
    # upgrades from 650/653/656.
    if 'org.laptop.WebActivity' not in \
           [act.get_bundle_id() for act in get_activities()]:
        try:
            d = os.path.dirname( USER_GROUPS_FILE )
            if not os.path.isdir(d):
                os.makedirs(d)
            with open( USER_GROUPS_FILE, 'w') as f:
                print >>f, DEFAULT_G1G1_GROUP
        except (IOError, OSError):
            pass # don't let this kill us! dlo trac #7845, #8681
        return [ DEFAULT_G1G1_GROUP ]
    # otherwise, there really isn't a group.  Ho, hum.
    return []
    
def get_activities():
    """Return a list of all the activities on the system."""
    # XXX: we should use the ActivityRegistry, but it needs the gtk main
    #      loop, which we don't want to deal with at the moment.
    actlist = []
    for dname in (USER_ACTIVITY_DIR, SYSTEM_ACTIVITY_DIR):
        if not os.path.isdir(dname): continue # skip missing dirs
        for fname in os.listdir(dname):
            path = os.path.join(dname, fname)
            if not os.path.exists(os.path.join(path,'activity','activity.info')):
                continue # skip this, not really an activity.
            try:
                a = actbun.ActivityBundle(path)
                # XXX: HACK.  Add some useful methods to this object.
                def maker(ab, func):
                    return lambda : func(ab)
                a.get_icon_data=maker(a, _get_icon_data)
                a.get_update_url = maker(a, lambda self: self._update_url)
                _lookup_update_url(a, is_activity=True)
                actlist.append(a)
            except:
                traceback.print_exc() # I guess this wasn't a legit activity.
    return actlist

def get_libraries():
    """Return a list of all the library content bundles on the system."""
    # XXX: we should use the ActivityRegistry, but it needs the gtk main
    #      loop, which we don't want to deal with at the moment.
    liblist = []
    for dname in (USER_LIBRARY_DIR, SYSTEM_LIBRARY_DIR):
        if not os.path.isdir(dname): continue # skip missing dirs
        for fname in os.listdir(dname):
            path = os.path.join(dname, fname)
            if not os.path.exists(os.path.join(path,'library','library.info')):
                continue # skip this, not really a contentbundle.
            try:
                l = libbun.ContentBundle(path)
                # XXX: HACK.  Add some useful methods to this object.
                def maker(ab, func):
                    return lambda : func(ab)
                l.get_icon_data=lambda: DEFAULT_LIBRARY_ICON
                l.get_update_url = maker(l, lambda self: self._update_url)
                _lookup_update_url(l, is_activity=False)
                # abuse duck typing to make content bundle look like
                # an activity bundle
                l.get_activity_version = maker(l, lambda self: self.get_library_version())
                l.get_bundle_id = maker(l, lambda self: self._global_name)
                _lookup_global_name(l)
                liblist.append(l)
            except:
                traceback.print_exc() # I guess this wasn't a legit bundle
    return liblist
