#!/usr/bin/python
# Copyright (C) 2008 One Laptop Per Child Association, Inc.
# Licensed under the terms of the GNU GPL v2 or later; see COPYING for details.
# Written by C. Scott Ananian <cscott@laptop.org>
"""Tools for working with remote activity bundles."""

import os

from ConfigParser import ConfigParser
from StringIO import StringIO
from zipfile import ZipFile

import bitfrost.update.actinfo as actinfo
import bitfrost.util.urlrange as urlrange
from sugar.bundle.activitybundle import ActivityBundle
from sugar.bundle.contentbundle import ContentBundle

def activity_info_from_url(url):
    """Return a `ConfigParser` for the activity.info file inside an .xo
    activity bundle at the location specified by `url`.  Attempts to use
    byte-range support on the target HTTP server to do this efficiently."""
    zf = ZipFile(urlrange.urlopen(url))
    return activity_info_from_zipfile(zf)

def activity_info_from_zipfile(zf):
    """Return a `ConfigParser` for the activity.info file inside the .xo
    activity bundle given by `zf`, which is a `ZipFile`."""
    cp = ConfigParser()
    filename = '%s/activity/activity.info' % bundle_base_from_zipfile(zf)
    cp.readfp(StringIO(zf.read(filename)))
    return cp

def library_info_from_zipfile(zf):
    """Return a `ConfigParser` for the library.info file inside the .xol
    content bundle given by `zf`, which is a `ZipFile`."""
    cp = ConfigParser()
    filename = '%s/library/library.info' % bundle_base_from_zipfile(zf)
    cp.readfp(StringIO(zf.read(filename)))
    return cp

def locale_activity_info_from_zipfile(zf):
    """Return a `ConfigParser` for the localized activity.linfo file inside
    the .xo activity bundle given by `zf`, which is a `ZipFile`."""
    import locale
    lang = locale.getdefaultlocale()[0]
    if lang is None: return None # bail early
    bundle_base = bundle_base_from_zipfile(zf)
    for f in ['locale/%s/activity.linfo' % lang,
              'locale/%s/activity.linfo' % lang[:2]]:
        filename = '%s/%s' % (bundle_base, f)
        if filename not in zf.namelist(): continue
        cp = ConfigParser()
        cp.readfp(StringIO(zf.read(filename)))
        return cp
    return None # can't find a localized activity.info

def locale_section_for_content_bundle(cp):
    """Return the proper section for a ConfigParser corresponding to a
    library.info file."""
    import locale
    lang = locale.getdefaultlocale()[0]
    if lang is None: return None # bail early
    for s in (lang, lang[:2]):
        if cp.has_section(s):
            return s
    return None # can't find a localized library.info section

def bundle_base_from_zipfile(zf):
    """Return the bundle directory base for the activity or content bundle given
    by the given `ZipFile` `zf`."""
    for TRAILING_PART in ('/activity/activity.info', '/library/library.info'):
        n = [n for n in zf.namelist() if
             n.endswith(TRAILING_PART) and n.count('/') == 2]
        if len(n)==1:
            return n[0][:-len(TRAILING_PART)]
    # hrm, this bundle is bogus.
    raise RuntimeError("Bad activity or content bundle: couldn't find " \
                       "activity.info or library.info")
    
def id_and_version_from_url(url):
    """Given the URL of a .xo file, return the activity id and
    integer version number, or (None,None) on error.

    Attempts to use byte-range support on the rarget HTTP server to extract
    just the activity/activity.info file and parse it.
    """
    try:
        cp = activity_info_from_url(url)
        SECTION = 'Activity'
        version = cp.get(SECTION, 'activity_version')
        if cp.has_option(SECTION, 'bundle_id'):
            bundle_id = cp.get(SECTION, 'bundle_id')
        else:
            bundle_id = cp.get(SECTION, 'service_name')
        return (bundle_id, version)
    except:
        return None, None

class BundleHelper(object):
    """This class helps with the installation/upgrade of activity and
    content bundles."""
    def __init__(self, filename):
        # check for a activity.info file to determine whether this is an
        # activity or content bundle.
        zf = ZipFile(filename)
        bundle_base = bundle_base_from_zipfile(zf)
        if ('%s/activity/activity.info' % bundle_base) in zf.namelist():
            self.bundle = ActivityBundle(filename)
            self.is_activity = True
        else:
            self.bundle = ContentBundle(filename)
            self.is_activity = False

    def is_installed(self, registry):
        """Return True if (an older version of) the bundle is already
        installed."""
        if self.is_activity:
            return registry.is_installed(self.bundle)
        else:
            return self.bundle.is_installed()

    def get_name(self):
        """Get the name of the activity or content from the bundle."""
        return self.bundle.get_name()

    def get_icon_data(self):
        """Get the icon data from the bundle."""
        if self.is_activity:
            # XXX should use get_icon_data()
            iconname = self.bundle.get_icon()
            try:
                return open(iconname).read()
            finally:
                os.unlink(iconname)
        else:
            return actinfo.DEFAULT_LIBRARY_ICON

    def _activity_install_or_upgrade(self, registry):
        """Install/upgrade an activity bundle, preserving favorites state."""
        if self.is_installed(registry):
            # upgraded activities are favorites if the pre-upgrade
            # activity was a favorite.
            ainfo = registry.get_activity(self.bundle.get_bundle_id())
            favorite = ainfo.favorite
            registry.upgrade(self.bundle)
        else:
            # newly installed activities will be favorites
            favorite = True
            registry.install(self.bundle)
        # set favorites flag for the new bundle-id/version.
        registry.set_bundle_favorite(self.bundle.get_bundle_id(),
                                     self.bundle.get_activity_version(),
                                     favorite)
    def _library_install_or_upgrade(self, registry):
        """Install/upgrade a content bundle."""
        if self.is_installed(registry):
            self.bundle.uninstall()
        self.bundle.install()

    def install_or_upgrade(self, registry):
        """Install/upgrade the activity or content bundle.  The `registry`
        parameter should be an instance of `sugar.activity.registry`."""
        if self.is_activity:
            self._activity_install_or_upgrade(registry)
        else:
            self._library_install_or_upgrade(registry)
