#!/usr/bin/python
# Copyright (C) 2008 One Laptop Per Child Association, Inc.
# Licensed under the terms of the GNU GPL v2 or later; see COPYING for details.
# Written by C. Scott Ananian <cscott@laptop.org>
"""Activity information microformat parser.

Activity information is embedded in HTML/XHTML/XML pages using a
semantic `microformat <http://microformats.org/>`_.  The following
CSS rules pull out the desired information.

Repository/group information for related activities is denoted by the
`CSS <http://www.w3.org/Style/CSS/>`_ selector::

  #olpc-activity-group-name

which provides a "short name" for the group, and::

  #olpc-activity-group-desc

which provides a slightly longer description.  As indicated by the fact
that these are matches on the 'id' attribute, they should match exactly
once.  These attributes are optional, and are not expected to be included
in pages describing updates for an individual activity. 

Each block of activity information is denoted by the selector::

  .olpc-activity-info

All information within that block describes a single available
version of an activity, so the following selectors should match
at most once within the block.

The activity id, which should be unique among all activities, is denoted
by the contents of the element identified by the selector::

  .olpc-activity-info .olpc-activity-id

The version number is denoted by the contents of the element identified by
the selector::

  .olpc-activity-info .olpc-activity-version

The version URL is denoted by the contents of the href attribute of the
element identified by the selector::

  .olpc-activity-info .olpc-activity-url *[href]

An example::

  <table class="olpc-activity-info">
   <tr>
    <td>Browse</td>
    <td class="olpc-activity-id"
        style="display:none;">org.laptop.WebActivity</td>
    <td class="olpc-activity-version">54</td>
    <td class="olpc-activity-url"><a href="Browse-54.xo">download!</a></td>
   </tr>
  </table>

"""
from __future__ import with_statement
from __future__ import division

from HTMLParser import HTMLParser, HTMLParseError
from sugar.bundle.bundleversion import NormalizedVersion
from sugar.bundle.bundleversion import InvalidVersionError

import urlparse

from bitfrost import _assert

_DEBUG_PARSER = False
"""Check that the HTML input is well-formed.
Otherwise, silently "do our best" with what we're given."""

class _UpdateHTMLParser(HTMLParser):
    """HTML parser to pull out data expressed in our microformat."""
    def _unrelative(self, url):
        if self.base_href is None: return url # no worries.
        return urlparse.urljoin(self.base_href, url)
    def __init__(self, base_href=None, try_hard=False):
        """The `base_href` parameter helps resolve relative URLS; the
        `try_hard` parameter tries to make up for deficient activity
        records with additional network queries."""
        HTMLParser.__init__(self)
        self.base_href = base_href
        self.try_hard = try_hard
        self.results = {}
        self.group_name = self.group_desc = None
        self.in_group_name = self.in_group_desc = self.in_activity = 0
        self._clear_info()
    def _clear_info(self):
        self.in_activity_id=self.in_activity_url=self.in_activity_version = 0
        self.last_id = self.last_version = None
        self.last_urls = []
    def handle_starttag(self, tag, attrs):
        classes = ' '.join([val for attr, val in attrs if attr=='class'])\
                  .split()
        if self.in_group_name == 0:
            if ('id','olpc-activity-group-name') in attrs:
                self.in_group_name = 1
        else:
            self.in_group_name += 1
        if self.in_group_desc == 0:
            if ('id','olpc-activity-group-desc') in attrs:
                self.in_group_desc = 1
        else:
            self.in_group_desc += 1
        if self.in_activity == 0:
            if 'olpc-activity-info' in classes:
                self.in_activity = 1
                self._clear_info()
            return
        self.in_activity += 1    
        if self.in_activity_id == 0:
            if 'olpc-activity-id' in classes:
                self.in_activity_id = 1
        else:
            self.in_activity_id += 1
        if self.in_activity_version == 0:
            if 'olpc-activity-version' in classes:
                self.in_activity_version = 1
        else:
            self.in_activity_version += 1
        if self.in_activity_url == 0:
            if 'olpc-activity-url' in classes:
                self.in_activity_url = 1
        else:
            self.in_activity_url += 1
        # an href inside activity_url is the droid we are looking for.
        if self.in_activity_url > 0:
            self.last_urls += [self._unrelative(v) for
                               a,v in attrs if a=='href']
    def handle_data(self, data):
        if self.in_group_name:
            self.group_name = data.strip()
        if self.in_group_desc:
            self.group_desc = data.strip()
        if self.in_activity_id > 0:
            self.last_id = data.strip()
        if self.in_activity_version > 0:
            try:
		self.last_version = NormalizedVersion(data.strip())
	    except InvalidVersionError:
                if _DEBUG_PARSER:
                    print "BAD VERSION NUMBER:", self.last_id, data

    def handle_endtag(self, tag):
        if self.in_group_name > 0:
            self.in_group_name -= 1
        if self.in_group_desc > 0:
            self.in_group_desc -= 1
        if self.in_activity_id > 0:
            self.in_activity_id -= 1
        if self.in_activity_version > 0:
            self.in_activity_version -= 1
        if self.in_activity_url > 0:
            self.in_activity_url -= 1
        if self.in_activity > 0:
            self.in_activity -= 1
            if self.in_activity == 0:
                if _DEBUG_PARSER:
                    _assert(self.in_activity_id == 0)
                    _assert(self.in_activity_version == 0)
                    _assert(self.in_activity_url == 0)
                if self.last_id is not None and self.last_id.strip() == '':
                    self.last_id = None
                if len(self.last_urls)>0 and (self.last_id is None or
                                              self.last_version is None):
                    # grumble.  look inside the URL to get the id/version
                    if self.try_hard:
                        from actutils import id_and_version_from_url
                        self.last_id, self.last_version = \
                                      id_and_version_from_url(self.last_urls[0])
                if self.last_version is not None and self.last_id is not None\
                   and len(self.last_urls)>0:
                    rl = self.results.get(self.last_id, [])
                    for url in self.last_urls:
                        if (self.last_version, url) not in rl:
                            rl.append((self.last_version, url))
                    self.results[self.last_id] = rl


def only_best_update(version_list):
    """Take a list of version, url pairs (like those returned from
    `parse_html` and return the "best" one."""
    bestv, bestu = None, None
    for ver, url in version_list:
        if bestv is None or ver > bestv:
            bestv, bestu = ver, url
    return bestv, bestu

def parse_html(html_data, base_href=None):
    """Parse the activity information embedded in the given string
    containing HTML data.  Returns a triple: the activity short name,
    if present (else `None`), then the activity description if present
    (else `None`), and finally a mapping from activity ids to a list
    of (version, url) pairs.

    Raises `HTMLParseError` if `html_data` can't be parsed as valid HTML.
    """
    hp = _UpdateHTMLParser(base_href)
    hp.feed(html_data)
    hp.close()
    return hp.group_name, hp.group_desc, hp.results

def parse_url(url, **urlopen_args):
    """Parse the activity information at the given URL. Returns the same
    information as `parse_html` does, and raises the same exceptions.
    The `urlopen_args` can be any keyword arguments accepted by
    `bitfrost.util.urlrange.urlopen`."""
    import bitfrost.util.urlrange as urlrange
    with urlrange.urlopen(url, **urlopen_args) as f:
        return parse_html(f.read(), f.url)
